# Autor: Mario Castañeda González
# TFG: Evaluación en línea del grado de involucración del usuario en actividades de interacción humano-robot

import cv2
import numpy as np

class Shoulder:
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

def head_pose_zone(image, face_mesh):
    # Voltea la imagen horizontalmente y convierte del espacio BGR (opencv) a RGB (mediapipe)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    # Impedimos que la imagen sea editable (mejora rendimiento)
    image.flags.writeable = False

    # Obtenemos malla con 468 puntos clave de la cara
    results = face_mesh.process(image)

    # Volvemos a permitir edición sobre la imagen para convertir de nuevo a BGR y trabajar con opencv
    image.flags.writeable = True
    image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)

    # Alto, ancho, nº canales de la imagen
    img_h, img_w, img_c = image.shape

    nose_2d = None

    if results.multi_face_landmarks:
        for face_landmarks in results.multi_face_landmarks:
            face_3d = []
            face_2d = []
            for idx, lm in enumerate(face_landmarks.landmark):
                # 33 - Comisura ojo derecho
                # 263 - Comisura ojo izquierdo
                # 1 - Punta de la nariz
                # 61 - Comisura derecha labio
                # 291 - Comisura izquierda labio
                # 199 - Barbilla
                if idx == 33 or idx == 263 or idx == 1 or idx == 61 or idx == 291 or idx == 199:
                    # Almacenamos datos de la nariz para trabajar más tarde sobre este punto
                    if idx == 1:
                        nose_2d = (lm.x * img_w, lm.y * img_h)

                    # Multiplicamos por el alto y el ancho de la imagen para escalar los resultados
                    x, y = int(lm.x * img_w), int(lm.y * img_h)

                    # Almacenamos coordenadas 2D de los puntos seleccionados
                    face_2d.append([x, y])

                    # Almacenamos coordenadas 3D de los puntos seleccionados
                    face_3d.append([x, y, lm.z])

                if idx == 13: labio_sup = np.array([int(lm.x * img_w), int(lm.y * img_h), lm.z])
                elif idx == 14: labio_inf = np.array([int(lm.x * img_w), int(lm.y * img_h), lm.z])
                elif idx == 61: com_der = np.array([int(lm.x * img_w), int(lm.y * img_h), lm.z])
                elif idx == 291: com_izq = np.array([int(lm.x * img_w), int(lm.y * img_h), lm.z])

            # Calculamos la apertura de la boca
            calc = labio_sup - labio_inf
            open_mouth = np.sqrt(np.dot(calc.T, calc))

            # Convertimos las coordenadas a un tipo de dato que maneje numpy
            face_2d = np.array(face_2d, dtype=np.float64)
            face_3d = np.array(face_3d, dtype=np.float64)

            # Distancia focal
            focal_length = 1 * img_w

            # Matriz de la cámara
            cam_matrix = np.array([[focal_length, 0, img_h / 2],
                                   [0, focal_length, img_w / 2],
                                   [0, 0, 1]])

            # Matriz de distorsión
            dist_matrix = np.zeros((4, 1), dtype=np.float64)

            # Aplicar PnP
            success, rot_vec, trans_vec = cv2.solvePnP(face_3d, face_2d, cam_matrix, dist_matrix)

            # Obtener matriz de rotación a partir del vector
            rmat, jac = cv2.Rodrigues(rot_vec)

            # Obtener ángulos
            angles, mtxR, mtxQ, Qx, Qy, Qz = cv2.RQDecomp3x3(rmat)

            # Convertimos la rotación a grados sexagesimales
            y = angles[0] * 360
            x = angles[1] * 360
            z = angles[2] * 360

            # Mostrar dirección nariz
            p1 = (int(nose_2d[0]), int(nose_2d[1]))
            p2 = (int(nose_2d[0] + x * 5), int(nose_2d[1] - y * 5))

        return x, y, z, [p1, p2], open_mouth
    else:
        return None, None, None, [None, None], None


def body_pose_zone(image, pose, dist_x_nose):
    # Voltea la imagen horizontalmente y convierte del espacio BGR (opencv) a RGB (mediapipe)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    # Impedimos que la imagen sea editable (mejora rendimiento)
    image.flags.writeable = False

    # Obtenemos malla con 32 puntos clave del cuerpo
    results = pose.process(image)

    # Volvemos a permitir edición sobre la imagen para convertir de nuevo a BGR y trabajar con opencv
    image.flags.writeable = True
    image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)

    # Alto, ancho, nº canales de la imagen
    img_h, img_w, img_c = image.shape

    # Creamos variables para evitar errores
    left_shoulder = None
    right_shoulder = None

    # La cámara solo obtiene 2 puntos de torso por la distancia hasta nuestro objetivo
    if results.pose_landmarks:
        for idx, lm in enumerate(results.pose_landmarks.landmark):
            if idx == 11:
                # left_shoulder = Shoulder(int(lm.x * img_w), int(lm.y * img_h), lm.z)
                left_shoulder = Shoulder(lm.x, lm.y, lm.z)
            elif idx == 12:
                # right_shoulder = Shoulder(int(lm.x * img_w), int(lm.y * img_h), lm.z)
                right_shoulder = Shoulder(lm.x, lm.y, lm.z)

    if not left_shoulder or not right_shoulder:
        print("No existe información de los hombros")
        return None, [None, None]
    else:
        # Rot respecto al eje Y
        modulo_planoXZ = np.sqrt((left_shoulder.z - right_shoulder.z) ** 2 + (left_shoulder.x - right_shoulder.x) ** 2)
        dif_Z = left_shoulder.z - right_shoulder.z
        rot_Y = (np.arcsin(dif_Z / modulo_planoXZ) * 180 / np.pi)

        # Mostrar dirección nariz
        p1 = (int((left_shoulder.x + right_shoulder.x) * img_w / 2),
              int((left_shoulder.y + right_shoulder.y) * img_h / 2))
        # p2 = (int(((left_shoulder.x + right_shoulder.x) * img_w / 2) + rot_Y * 3  - (dist_x_nose/2)),
        p2 = (int(((left_shoulder.x + right_shoulder.x) * img_w / 2) + rot_Y * 3),
              int((left_shoulder.y + right_shoulder.y) * img_h / 2))

        # cv2.imshow('Posición torso', image)
        return rot_Y, [p1, p2]

def aprox_nose(pos_nose, last_nose):
    if last_nose[0] - 50 < pos_nose[0] < last_nose[0] + 50 and last_nose[1] - 50 < pos_nose[1] < last_nose[1] + 50:
        return True
    else:
        return False